import _typeof from 'babel-runtime/helpers/typeof';
import _Object$getPrototypeOf from 'babel-runtime/core-js/object/get-prototype-of';
import _Object$create from 'babel-runtime/core-js/object/create';
import invariant from 'invariant';
import searchParentConstructorEntry from './searchParentConstructorEntry';

var emptySymbol = 'empty';
var unwrappedMethods = ['constructor', 'get', 'getIn', 'first', 'last', 'reduce', 'reduceRight', 'find', 'findLast', 'findEntry', 'findLastEntry', 'max', 'maxBy', 'min', 'minBy', 'clear' // Important! We're manually overriding this method
];

export default function createExtendable(base, copy, empty) {
  invariant(typeof copy === 'function', name + ': `copy` is expected to be a function.');
  invariant(typeof empty === 'function', name + ': `empty` is expected to be a function.');

  var constructor = base.prototype.constructor;
  var name = constructor.name;
  var proto = _Object$create(base.prototype);

  // Overrides the original clear method that returns an empty object
  proto.clear = function clear() {
    return this.__wrapImmutable({});
  };

  // A method for wrapping an immutable object, with reference equality for empty objects
  proto.__wrapImmutable = function __wrapImmutable(val) {
    var forceCreation = arguments.length <= 1 || arguments[1] === undefined ? false : arguments[1];

    var prototype = _Object$getPrototypeOf(this);
    var constructor = prototype.constructor;


    if (!val.size && !val.__ownerID && !forceCreation) {
      if (!constructor[emptySymbol] || searchParentConstructorEntry(constructor, emptySymbol, constructor[emptySymbol])) {
        return constructor[emptySymbol] = empty(_Object$create(prototype));
      }

      return constructor[emptySymbol];
    }

    return copy(_Object$create(prototype), val);
  };

  // Methods which will yield a Map and have to be wrapped before returning a result
  for (var key in base.prototype) {
    if (!key.startsWith('__') && !key.startsWith('to') && unwrappedMethods.indexOf(key) === -1) {
      (function () {
        var _originalMethod = base.prototype[key];

        if (typeof _originalMethod === 'function') {
          proto[key] = function wrappedMethod() {
            for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
              args[_key] = arguments[_key];
            }

            var res = _originalMethod.apply(this, args);
            var _constructor = void 0;

            if (res && (typeof res === 'undefined' ? 'undefined' : _typeof(res)) === 'object' && (_constructor = _Object$getPrototypeOf(res).constructor) === constructor) {
              return this.__wrapImmutable(res);
            }

            return res;
          };
        }
      })();
    }
  }

  proto.__ensureOwner = function __ensureOwner(ownerID) {
    if (ownerID === this.__ownerID) {
      return this;
    } else if (!ownerID) {
      this.__ownerID = undefined;
      this.__altered = false;
      return this;
    }

    var res = this.__wrapImmutable(this, true);
    res.__ownerID = ownerID;
    return res;
  };

  return proto;
}